/*
   Copyright (C) 2009 Roman Belov

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <math.h>

#include <SDL/SDL.h>
#undef	main

#include "caph.h"
#include "draw.h"
#include "phys.h"

#if defined(_WIN32)
# include <windows.h>
#else
# include <unistd.h>
#endif

#define SYS_DATA_DIR	"../data/"

static void 
sys_chdir(const char *name)
{
#if defined(_WIN32)
	SetCurrentDirectory(name);
#else
	chdir(name);
#endif
}

static void
sys_sleep(long msec)
{
#if defined(_WIN32)
	Sleep(msec);
#else
	usleep(msec * 1000);
#endif
}

int		screen_w;
int		screen_h;
uint8_t		*screen_pixels;

#define	STRIP_SIZE		(512)

static point_t  *p_strip;
static int	p_it;
static int	p_state;
static int	p_flags;
static int	time;

static void
phys_paint_click(float x, float y, int state)
{
	concave_t	*cx;

	if (p_state && !state) {

		if (p_it > 2) {
			cx = phys_alloc(p_it + 1);
			cx->flags = p_flags;
			memcpy(cx->pts, p_strip, sizeof(point_t) * cx->pt);
			phys_insert(cx);
		}

		p_it = 0;
	}

	if (!p_state && state) {
		p_strip[p_it].x = x;
		p_strip[p_it].y = y;
	}

	p_state = state;
}

static void
phys_paint_move(float x, float y)
{
	float		dist;
	float		dx, dy;

	if (!p_state)
		return;

	dx = x - p_strip[p_it].x;
	dy = y - p_strip[p_it].y;

	dist = sqrt(dx * dx + dy * dy);

	if (dist < 8.0f)
		return;

	if (p_it > 1 && dist < 36.0f) {

		float		pdx, pdy;

		pdx = p_strip[p_it].x - p_strip[p_it - 1].x;
		pdy = p_strip[p_it].y - p_strip[p_it - 1].y;

		dist = sqrt(fabsf(pdx * dy - pdy * dx));

		if (dist < 8.0f)
			return;
	}

	if (p_it < (STRIP_SIZE - 1)) {
		p_strip[++p_it].x = x;
		p_strip[p_it].y = y;
	}
}

static void
phys_paint_draw()
{
	color_t		col = { 85, 85, 85, 0 };
	int		i;

	if (p_it > 1) {
		for (i = 1; i <= p_it; ++i) {
			draw_line(p_strip + i - 1,
					p_strip + i, col);
		}
	}
}

static void
phys_flags_draw()
{
	printf("flags: %s%s%s%s \t color: %s%s%s \t state: %s\n",
			(p_flags & CONCAVE_CONST) ? "const" : "",
			(p_flags & CONCAVE_SOFT) ? "soft" : "",
			(p_flags & CONCAVE_BEND) ? "bend" : "",
			(p_flags & CONCAVE_PHANTOM) ? "phantom" : "",
			(p_flags & CONCAVE_PHANTOM_RED) ? "r" : "_",
			(p_flags & CONCAVE_PHANTOM_GREEN) ? "g" : "_",
			(p_flags & CONCAVE_PHANTOM_BLUE) ? "b" : "_",
			(time) ? "run" : "stop");
}

#define DATA_CONFIG		"caph.conf"
#define DATA_PAPER		"paper.png"

int main(int argc, char *argv[])
{
	SDL_Surface *screen;
	SDL_Event event;
	int fs = 0;
	int run = 1;

	sys_chdir(SYS_DATA_DIR);

	FILE *conf = fopen(DATA_CONFIG, "r");

	if (conf == NULL) {
		fprintf(stderr, "%s:%i [PANIC] fopen(\"%s\") failed: %s \n",
				__FILE__, __LINE__, DATA_CONFIG, strerror(errno));
		exit(EXIT_FAILURE);
	}

	fscanf(conf, "%i %i %i",
			&screen_w,
			&screen_h,
			&fs);

	fclose(conf);

	if (screen_w < 320 || screen_w > (1024*4) ||
			screen_h < 240 || screen_h > (1024*4))
	{
		fprintf(stderr, "%s:%i [PANIC] Ivalid config values\n",
				__FILE__, __LINE__);
	}

	if (SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "%s:%i [PANIC] SDL_Init failed: %s \n",
				__FILE__, __LINE__, SDL_GetError());
		exit(EXIT_FAILURE);
	}

	int sdl_flags = (fs ? SDL_FULLSCREEN : 0) |
			SDL_DOUBLEBUF |
			SDL_HWSURFACE;

#ifdef _OPENGL
	SDL_GL_SetAttribute(SDL_GL_BUFFER_SIZE, 24);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 0);
	SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 0);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	sdl_flags |= SDL_OPENGL;
#endif

	screen = SDL_SetVideoMode(screen_w, screen_h, 32, sdl_flags);

	if (screen == NULL) {
		fprintf(stderr, "%s:%i [PANIC] SDL_SetVideoMode failed: %s \n",
				__FILE__, __LINE__, SDL_GetError());
		SDL_Quit();
		exit(EXIT_FAILURE);
	}

	screen_pixels = screen->pixels;

	draw_init();
	draw_load_bg(DATA_PAPER);

	p_strip = malloc(sizeof(point_t) * STRIP_SIZE);
	p_it = 0;
	p_state = 0;
	p_flags = 0;

	phys_open();

	concave_t *cx = phys_alloc(5);
	cx->flags = CONCAVE_CONST;
	cx->pts[0].x = 0.0f;
	cx->pts[0].y = 0.0f;
	cx->pts[1].x = 0.0f;
	cx->pts[1].y = (float) (screen_h - 1);
	cx->pts[2].x = (float) (screen_w - 1);
	cx->pts[2].y = (float) (screen_h - 1);
	cx->pts[3].x = (float) (screen_w - 1);
	cx->pts[3].y = 0.0f;
	cx->pts[4].x = 0.0f;
	cx->pts[4].y = 0.0f;

	phys_insert(cx);
	phys_remove_hold();

	time = 1;

        int dt, t = 0;
        int fr_n = 0;
        int fr_tval = 0;
        int fps = 0;
        int dt_min = (1000/60);
	uint32_t t0, t1;

	t0 = t1 = SDL_GetTicks();

	do {
		while (SDL_PollEvent(&event)) {
			switch (event.type) {
				case SDL_QUIT:
					run = 0;
					break;

				case SDL_KEYDOWN:
					switch ((int) event.key.keysym.sym) {
						case SDLK_q:
							run = 0;
							break;

						case SDLK_a:
							p_flags &= ~CONCAVE_TYPE_MASK;
							break;

						case SDLK_c:
							p_flags &= ~CONCAVE_TYPE_MASK;
							p_flags |= CONCAVE_CONST;
							break;

						case SDLK_s:
							p_flags &= ~CONCAVE_TYPE_MASK;
							p_flags |= CONCAVE_SOFT;
							break;

						case SDLK_b:
							p_flags &= ~CONCAVE_TYPE_MASK;
							p_flags |= CONCAVE_BEND;
							break;

						case SDLK_x:
							p_flags &= ~CONCAVE_TYPE_MASK;
							p_flags |= CONCAVE_PHANTOM;
							break;

						case SDLK_1:
							p_flags ^= CONCAVE_PHANTOM_RED;
							break;

						case SDLK_2:
							p_flags ^= CONCAVE_PHANTOM_GREEN;
							break;

						case SDLK_3:
							p_flags ^= CONCAVE_PHANTOM_BLUE;
							break;

						case SDLK_d:
							phys_remove_back();
							break;

						case SDLK_t:
							time ^= 1;
							break;
					}
					phys_flags_draw();
					break;

				case SDL_MOUSEBUTTONDOWN:
				case SDL_MOUSEBUTTONUP:
					if (event.button.button == SDL_BUTTON_LEFT) {
						phys_paint_click((float) event.button.x,
								(float) event.button.y,
								event.type == SDL_MOUSEBUTTONDOWN);
					}
					if (event.button.button == SDL_BUTTON_RIGHT) {
						if (event.type == SDL_MOUSEBUTTONDOWN) {
							phys_pick((float) event.button.x,
									(float) event.button.y);
						}
						else
							phys_pick_detach();
					}
					break;

				case SDL_MOUSEMOTION:
					phys_paint_move((float) event.motion.x,
							(float) event.motion.y);
					phys_pick_move((float) event.button.x,
							(float) event.button.y);
					break;
			}
		}

		t0 = SDL_GetTicks();
		dt = t0 - t1;

		while (dt < dt_min) {
			sys_sleep(1);
			t0 = SDL_GetTicks();
			dt = t0 - t1;
		}

		t1 = t0;

		++fr_n;
		t += dt;
		fr_tval += dt;

		if (fr_tval > 1000) {
			fps = (fr_n * 1000) / fr_tval;
			printf("fps: %i \n", fps);
			fflush(stdout);
			fr_n = 0;
			fr_tval = 0;
		}

		if (time) {
			phys_update((float) dt * 0.001f);
		}

		if (SDL_LockSurface(screen) == 0) {

			draw_clear();

			phys_draw();
			phys_paint_draw();

			SDL_UnlockSurface(screen);
#ifdef _OPENGL
			SDL_GL_SwapBuffers();
#else
			SDL_Flip(screen);
#endif
		}

	} while (run);

	phys_close();

	SDL_Quit();
}

