/*
   Copyright (C) 2009 Roman Belov

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "phys.h"
#include "caph.h"

#define	PHYS_DELTA	0.0078125f

float			phys_g = 0.1f;

static concave_t	*all;
static float		adt;

#define	GLOB_LINKS_MAX	64

static link_t		*glob_lns;
static int		glob_ln;

#define	GLOB_PICK_AREA	16.0f

static point_t		*glob_pick_point;
static point_t		glob_pick_dst;
static int		glob_pick;

static concave_t	*glob_rm_hold;

static void
all_list_insert(concave_t *cx)
{
	if (all) {
		cx->next = all->next;
		all->next->prev = cx;
		cx->prev = all;
		all->next = cx;
	}
	else {
		all = cx;
		all->prev = all;
		all->next = all;
	}
}

static void
all_list_remove(concave_t *cx)
{
	if (all == all->next)
		all = NULL;
	else {
		if (all == cx)
			all = all->prev;
		cx->prev->next = cx->next;
		cx->next->prev = cx->prev;
	}
}

concave_t *phys_alloc(int len)
{
	concave_t	*cx;

	cx = malloc(sizeof(concave_t));
	cx->pt = len;
	cx->lns = NULL;

	cx->last = malloc(sizeof(point_t) * len);
	cx->pts = malloc(sizeof(point_t) * len);

	return cx;
}

void phys_free(concave_t *cx)
{
	free(cx->pts);
	free(cx->last);
	if (cx->lns) free(cx->lns);
	free(cx);
}

void phys_free_all()
{
	concave_t	*curr = all;
	concave_t	*next;

	if (!curr) return;

	do {
		next = curr->next;
		phys_free(curr);
		curr = next;
	}
	while (curr != all);

	all = NULL;
}

static int
phys_intersect_lines(
		const point_t *l1a,
		const point_t *l1b,
		const point_t *l2a,
		const point_t *l2b)
{
	point_t		n;
	float		d, sgn;

	n.y = (l1a->x - l1b->x);
	n.x = -(l1a->y - l1b->y);
	d = l1a->x * n.x + l1a->y * n.y;

	sgn = l2a->x * n.x + l2a->y * n.y - d;
	sgn *= l2b->x * n.x + l2b->y * n.y - d;

	if (sgn > 0.0f) return 0;

	n.y = (l2a->x - l2b->x);
	n.x = -(l2a->y - l2b->y);
	d = l2a->x * n.x + l2a->y * n.y;

	sgn = l1a->x * n.x + l1a->y * n.y - d;
	sgn *= l1b->x * n.x + l1b->y * n.y - d;

	if (sgn > 0.0f) return 0;

	return 1;
}

static void
phys_glob_link_put(point_t *p1, point_t *p2, int flags)
{
	float x, y;
	float dist;
	float w;

	if (glob_ln > (GLOB_LINKS_MAX - 1)) return;

	x = p1->x - p2->x;
	y = p1->y - p2->y;
	dist = sqrt(x * x + y * y);

	glob_lns[glob_ln].beg = p1;
	glob_lns[glob_ln].end = p2;
	glob_lns[glob_ln].dist = dist;

	if (flags & 1)
		w = 0.0f;
	else if (flags & 2)
		w = 1.0f;
	else
		w = 0.5f;

	glob_lns[glob_ln++].wght = w;
}

static void
phys_intersection_links(concave_t *cx, concave_t *px)
{
	point_t		*pts = cx->pts;
	int		pt = cx->pt;
	point_t		*p_pts = px->pts;
	int		p_pt = px->pt;
	int		i, j;
	int		flags, ret;

	flags =
		((cx->flags & CONCAVE_CONST) ? 1 : 0) |
		((px->flags & CONCAVE_CONST) ? 2 : 0) ;

	for (i = 1; i < pt; ++i) {
		for (j = 1; j < p_pt; ++j) {

			ret = phys_intersect_lines(
					pts + i - 1,
					pts + i,
					p_pts + j - 1,
					p_pts + j);

			if (ret) {
				phys_glob_link_put(
						pts + i - 1,
						p_pts + j - 1,
						flags);

				phys_glob_link_put(
						pts + i,
						p_pts + j,
						flags);

				phys_glob_link_put(
						pts + i - 1,
						p_pts + j,
						flags);

				phys_glob_link_put(
						pts + i,
						p_pts + j - 1,
						flags);
			}
		}
	}
}

static void
phys_intersection_links_all(concave_t *cx)
{
	concave_t	*curr = all;

	if (!curr) return;

	glob_ln = 0;

	do {
		do {
			if (curr == cx) break;

			if ((curr->flags & CONCAVE_CONST) &&
					(cx->flags & CONCAVE_CONST))
				break;

			if (curr->flags & CONCAVE_PHANTOM)
				break;

			phys_intersection_links(cx, curr);
		}
		while (0);

		curr = curr->next;
	}
	while (curr != all);
}

void phys_insert(concave_t *cx)
{
	float		x, y;
	float		dist;
	int		pt, ln;
	int		i, j;

	all_list_insert(cx);

	pt = cx->pt;
	ln = 0;

	/* Do not moving */

	for (i = 0; i < pt; ++i) {
		cx->last[i].x = cx->pts[i].x;
		cx->last[i].y = cx->pts[i].y;
	}

	/* Create joints */

	phys_intersection_links_all(cx);

	if (cx->flags & CONCAVE_CONST) {
		if (glob_ln) {
			cx->lns = malloc(sizeof(link_t) * (glob_ln));
		}
	}
	else if (cx->flags & CONCAVE_PHANTOM) {
		/* Do not save joints links */
		return ;
	}
	else if (cx->flags & CONCAVE_SOFT) {

		/* Wire-like objects */

		cx->lns = malloc(sizeof(link_t) * (pt - 1 + glob_ln));

		for (i = 1; i < pt; ++i) {
			x = cx->pts[i - 1].x - cx->pts[i].x;
			y = cx->pts[i - 1].y - cx->pts[i].y;
			dist = sqrt(x * x + y * y);

			cx->lns[ln].beg = cx->pts + i - 1;
			cx->lns[ln].end = cx->pts + i;
			cx->lns[ln].dist = dist;
			cx->lns[ln++].wght = 0.5f;
		}
	}
	else {
		/* Regular objects */

		int	ln_c = pt * (pt - 1) / 2;

		cx->lns = malloc(sizeof(link_t) * (ln_c + glob_ln));

		for (i = 0; i < pt; ++i) {
			for (j = i + 1; j < pt; ++j) {

				x = cx->pts[i].x - cx->pts[j].x;
				y = cx->pts[i].y - cx->pts[j].y;
				dist = sqrt(x * x + y * y);

				cx->lns[ln].beg = cx->pts + i;
				cx->lns[ln].end = cx->pts + j;
				cx->lns[ln].dist = dist;
				cx->lns[ln++].wght = 0.5f;
			}
		}
	}

	/* Copy joints to end */

	if (glob_ln) {
		memcpy(cx->lns + ln, glob_lns, sizeof(link_t) * glob_ln);
		ln += glob_ln;
	}

	cx->ln = ln;
}

void phys_remove(concave_t *cx)
{
	all_list_remove(cx);
	phys_free(cx);
}

void phys_remove_hold()
{
	glob_rm_hold = all->next;
}

void phys_remove_back()
{
	concave_t	*target = all->next;

	if (target != glob_rm_hold) {
		all_list_remove(target);
		phys_free(target);
	}
}

concave_t *phys_get_all()
{
	return all;
}

static void
phys_verlet(concave_t *cx)
{
	point_t		*last = cx->last;
	point_t		*pts = cx->pts;
	int		i, pt = cx->pt;
	float		x, y;

	for (i = 0; i < pt; ++i) {
		x = pts[i].x;
		y = pts[i].y;
		pts[i].x = x + (x - last[i].x);
		pts[i].y = y + (y - last[i].y) + phys_g;
		last[i].x = x;
		last[i].y = y;
	}
}

static void
phys_links(concave_t *cx)
{
	point_t		*beg, *end;
	link_t		*lns = cx->lns;
	int		ln = cx->ln;
	float		dx, dy;
	float		dist, diff, w;
	float		diff_beg;
	float		diff_end;
	int		i, it;

	if (cx->flags & CONCAVE_BEND) {
		for (it = 0; it < 8; ++it) {
			for (i = 0; i < ln; ++i) {
				beg = lns[i].beg;
				end = lns[i].end;
				dx = beg->x - end->x;
				dy = beg->y - end->y;
				dist = sqrt(dx * dx + dy * dy);
				dx /= dist;
				dy /= dist;
				diff = (dist - lns[i].dist);
				if (fabsf(diff) > 1.0f) {
					lns[i].dist += diff * 0.4f;
				}
				diff = (lns[i].dist - dist) * 0.5f;
				w = lns[i].wght;
				diff_beg = diff * w;
				diff_end = diff * (1.0f - w);
				beg->x += dx * diff_beg;
				beg->y += dy * diff_beg;
				end->x += -dx * diff_end;
				end->y += -dy * diff_end;
			}
		}
	}
	else {
		for (it = 0; it < 8; ++it) {
			for (i = 0; i < ln; ++i) {
				beg = lns[i].beg;
				end = lns[i].end;
				dx = beg->x - end->x;
				dy = beg->y - end->y;
				dist = sqrt(dx * dx + dy * dy);
				dx /= dist;
				dy /= dist;
				diff = (lns[i].dist - dist) * 0.5f;
				w = lns[i].wght;
				diff_beg = diff * w;
				diff_end = diff * (1.0f - w);
				beg->x += dx * diff_beg;
				beg->y += dy * diff_beg;
				end->x += -dx * diff_end;
				end->y += -dy * diff_end;
			}
		}
	}
}

#define PHYS_PUSH_BIAS		0.1f

static void
phys_push_point_line(
		point_t *pt,
		point_t *pt_p,
		point_t *la,
		point_t *lb,
		point_t *la_p,
		point_t *lb_p,
		int flags)
{
	point_t		n, s;
	float		dist, frct;
	float		d;

	n.y = (la->x - lb->x);
	n.x = -(la->y - lb->y);
	d = sqrt(n.x * n.x + n.y * n.y);
	n.x /= d;
	n.y /= d;
	d = la->x * n.x + la->y * n.y;

	dist = pt->x * n.x + pt->y * n.y - d;

	s.x = (pt->x - pt_p->x) - ((la->x - la_p->x) + (lb->x - lb_p->x)) * 0.5f;
	s.y = (pt->y - pt_p->y) - ((la->y - la_p->y) + (lb->y - lb_p->y)) * 0.5f;

	frct = s.x * n.y - s.y * n.x;

	if (flags & 4) {
		frct *= 0.1f;
	}

	if (flags & 2) {

		dist *= 1.5f;
		frct *= -0.4f;

		dist += (dist < 0.0f) ? - PHYS_PUSH_BIAS : + PHYS_PUSH_BIAS;

		pt->x +=	- n.x * dist	+ n.y * frct;
		pt->y +=	- n.y * dist	- n.x * frct;

		return;
	}

	if (flags & 1) {

		dist *= 1.5f;
		frct *= -0.4f;

		dist += (dist < 0.0f) ? - PHYS_PUSH_BIAS : + PHYS_PUSH_BIAS;

		la->x +=	+ n.x * dist	- n.y * frct;
		la->y +=	+ n.y * dist	+ n.x * frct;
		lb->x +=	+ n.x * dist	- n.y * frct;
		lb->y +=	+ n.y * dist	+ n.x * frct;

		return;
	}

	dist *= 0.75f;
	frct *= -0.2f;

	dist += (dist < 0.0f) ? - PHYS_PUSH_BIAS : + PHYS_PUSH_BIAS;

	pt->x +=	- n.x * dist	+ n.y * frct;
	pt->y +=	- n.y * dist	- n.x * frct;
	la->x +=	+ n.x * dist	- n.y * frct;
	la->y +=	+ n.y * dist	+ n.x * frct;
	lb->x +=	+ n.x * dist	- n.y * frct;
	lb->y +=	+ n.y * dist	+ n.x * frct;
}

#define	PHYS_EPS	0.00001f
#define NONZERO(x)	(fabsf(x) > PHYS_EPS)

static int
phys_ccd_point_line(
		point_t *pt,
		point_t *pt_p,
		point_t *la,
		point_t *lb,
		point_t *la_p,
		point_t *lb_p,
		int flags)
{
	point_t		pt_min, pt_max;
	point_t		ln_min, ln_max;
	point_t		pa, ba, ppaa, bbaa;
	point_t		n_pt, n_a, n_b;
	point_t		n, pb;
	float		a, b, c, d;
	float		t, t1, t2;

	/* Early test */

	/*n.x = pt->x - la->x;
	n.y = pt->y - la->y;
	d = n.x * n.x + n.y * n.y;

	if (d > (80.0f*80.0f)) return 0;*/

	/* AABB test */

	if (pt->x < pt_p->x) {
		pt_min.x = pt->x;
		pt_max.x = pt_p->x;
	}
	else {
		pt_min.x = pt_p->x;
		pt_max.x = pt->x;
	}

	if (pt->y < pt_p->y) {
		pt_min.y = pt->y;
		pt_max.y = pt_p->y;
	}
	else {
		pt_min.y = pt_p->y;
		pt_max.y = pt->y;
	}

	ln_min.x = la->x;
	if (lb->x < ln_min.x) ln_min.x = lb->x;
	if (la_p->x < ln_min.x) ln_min.x = la_p->x;
	if (lb_p->x < ln_min.x) ln_min.x = lb_p->x;

	if (pt_max.x < ln_min.x) return 0;

	ln_min.y = la->y;
	if (lb->y < ln_min.y) ln_min.y = lb->y;
	if (la_p->y < ln_min.y) ln_min.y = la_p->y;
	if (lb_p->y < ln_min.y) ln_min.y = lb_p->y;

	if (pt_max.y < ln_min.y) return 0;

	ln_max.x = la->x;
	if (lb->x > ln_max.x) ln_max.x = lb->x;
	if (la_p->x > ln_max.x) ln_max.x = la_p->x;
	if (lb_p->x > ln_max.x) ln_max.x = lb_p->x;

	if (pt_min.x > ln_max.x) return 0;

	ln_max.y = la->y;
	if (lb->y > ln_max.y) ln_max.y = lb->y;
	if (la_p->y > ln_max.y) ln_max.y = la_p->y;
	if (lb_p->y > ln_max.y) ln_max.y = lb_p->y;

	if (pt_min.y > ln_max.y) return 0;

	/* Full CCD test */

	pa.x = pt_p->x - la_p->x;
	pa.y = pt_p->y - la_p->y;

	ba.y = lb_p->x - la_p->x;
	ba.x = -(lb_p->y - la_p->y);

	ppaa.x = (pt->x - pt_p->x) - (la->x - la_p->x);
	ppaa.y = (pt->y - pt_p->y) - (la->y - la_p->y);

	bbaa.y = (lb->x - lb_p->x) - (la->x - la_p->x);
	bbaa.x = -((lb->y - lb_p->y) - (la->y - la_p->y));

	a = ppaa.x * bbaa.x + ppaa.y * bbaa.y;
	b = pa.x * bbaa.x + pa.y * bbaa.y + ba.x * ppaa.x + ba.y * ppaa.y;
	c = pa.x * ba.x + pa.y * ba.y;

	if (NONZERO(a)) {
		d = b * b - 4.0f * a * c;
		if (d < 0.0f) return 0;
		d = sqrt(d);
		t1 = (- b + d) / (2.0f * a);
		t2 = (- b - d) / (2.0f * a);
		if (t1 >= 0.0f && t1 <= 1.0f) {
			if (t2 >= 0.0f && t2 <= 1.0f)
				t = (t1 < t2) ? t1 : t2;
			else
				t = t1;
		}
		else if (t2 >= 0.0f && t2 <= 1.0f)
			t = t2;
		else
			return 0;
	}
	else if (NONZERO(b)) {
		t = - c / b;
		if (t < 0.0f || t > 1.0f) return 0;
	}
	else return 0;

	if (isnan(t)) return 0;

	n_pt.x = pt_p->x + (pt->x - pt_p->x) * t;
	n_pt.y = pt_p->y + (pt->y - pt_p->y) * t;

	n_a.x = la_p->x + (la->x - la_p->x) * t;
	n_a.y = la_p->y + (la->y - la_p->y) * t;

	n_b.x = lb_p->x + (lb->x - lb_p->x) * t;
	n_b.y = lb_p->y + (lb->y - lb_p->y) * t;

	n.x = n_b.x - n_a.x;
	n.y = n_b.y - n_a.y;

	pa.x = n_pt.x - n_a.x;
	pa.y = n_pt.y - n_a.y;
	d = pa.x * n.x + pa.y * n.y;

	pb.x = n_pt.x - n_b.x;
	pb.y = n_pt.y - n_b.y;
	d *= pb.x * n.x + pb.y * n.y;

	if (d > 0.0f) return 0;

	/* Collision detected, go to pushing */

	phys_push_point_line(pt, pt_p, la, lb, la_p, lb_p, flags);

	return 1;
}

static int glob_cd;

static void
phys_collise(const concave_t *cx, const concave_t *bound)
{
	point_t		*lst = cx->last;
	point_t		*pts = cx->pts;
	int		pt = cx->pt;
	point_t		*b_lst = bound->last;
	point_t		*b_pts = bound->pts;
	int		b_pt = bound->pt;
	int		i, j;
	int		flags, flags_swp;

	flags =
		((cx->flags & CONCAVE_CONST) ? 1 : 0) |
		((bound->flags & CONCAVE_CONST) ? 2 : 0) |
		(((cx->flags | bound->flags) & CONCAVE_SOFT) ? 4 : 0) ;

	flags_swp = (flags & 4) |
		((flags & 1) << 1) |
		((flags & 2) >> 1) ;

	for (i = 0; i < pt; ++i) {
		for (j = 1; j < b_pt; ++j) {

			glob_cd += phys_ccd_point_line(
					pts + i,
					lst + i,
					b_pts + j - 1,
					b_pts + j,
					b_lst + j - 1,
					b_lst + j,
					flags);
		}
	}

	for (i = 0; i < b_pt; ++i) {
		for (j = 1; j < pt; ++j) {

			glob_cd += phys_ccd_point_line(
					b_pts + i,
					b_lst + i,
					pts + j - 1,
					pts + j,
					lst + j - 1,
					lst + j,
					flags_swp);
		}
	}
}

static void
phys_collise_all(const concave_t *cx)
{
	concave_t	*curr = cx->next;

	while (curr != all) {

		do {
			if ((curr->flags & CONCAVE_CONST) &&
					(cx->flags & CONCAVE_CONST))
				break;

			if ((curr->flags & CONCAVE_SOFT) &&
					(cx->flags & CONCAVE_SOFT))
				break;

			if (curr->flags & CONCAVE_PHANTOM)
				break;

			phys_collise(cx, curr);
		}
		while (0);

		curr = curr->next;
	}
}

void phys_open()
{
	all = NULL;
	adt = 0.0f;

	glob_lns = malloc(sizeof(link_t) * GLOB_LINKS_MAX);
	glob_ln = 0;

	glob_pick = 0;
}

void phys_close()
{
	phys_free_all();
	free(glob_lns);
}

static void
phys_update_all()
{
	concave_t	*curr = all;

	if (!curr) return;

	do {
		do {
			if (curr->flags & CONCAVE_CONST)
				break;

			if (curr->flags & CONCAVE_PHANTOM)
				break;

			phys_verlet(curr);
			phys_links(curr);
		}
		while (0);

		curr = curr->next;
	}
	while (curr != all);

	if (glob_pick) {
		*glob_pick_point = glob_pick_dst;
	}

	int it = 4;

	do {
		curr = all;
		glob_cd = 0;

		do {
			do {
				if (curr->flags & CONCAVE_PHANTOM)
					break;

				phys_collise_all(curr);
			}
			while (0);

			curr = curr->next;
		}
		while (curr != all);

		--it;
	}
	while (glob_cd && it);
}

void phys_update(float dt)
{
	adt += dt;

	if (adt > (4.0f * PHYS_DELTA))
		adt = (4.0f * PHYS_DELTA);

	while (adt > PHYS_DELTA) {
		phys_update_all();
		adt -= PHYS_DELTA;
	}
}

static void
phys_draw_concave(concave_t *cx)
{
	color_t		col_def = { 25, 25, 25, 0 };
	color_t		col_const = { 155, 25, 25, 0 };
	color_t		col_soft = { 25, 25, 155, 0 };
	color_t		col_bend = { 25, 125, 25, 0 };
	color_t		col;
	point_t		*pts = cx->pts;
	int		i, pt = cx->pt;

	if (cx->flags & CONCAVE_CONST)
		col = col_const;
	else if (cx->flags & CONCAVE_SOFT)
		col = col_soft;
	else if (cx->flags & CONCAVE_BEND)
		col = col_bend;
	else if (cx->flags & CONCAVE_PHANTOM) {
		col.r = (cx->flags & CONCAVE_PHANTOM_RED) ? 200 : 50 ;
		col.g = (cx->flags & CONCAVE_PHANTOM_GREEN) ? 200 : 50 ;
		col.b = (cx->flags & CONCAVE_PHANTOM_BLUE) ? 200 : 50 ;
		col.a = 1;
	}
	else
		col = col_def;

	for (i = 1; i < pt; ++i) {
		draw_line(pts + i - 1,
				pts + i, col);
	}
}

void phys_draw()
{
	concave_t	*curr = all;

	if (!curr)
		return;

	do {
		phys_draw_concave(curr);
		curr = curr->prev;
	}
	while (curr != all);
}

static point_t *
phys_dist_point(concave_t *cx, float px, float py)
{
	point_t		*pts = cx->pts;
	int		pt = cx->pt;
	float		x, y;
	float		dist;
	int		i;

	for (i = 0; i < pt; ++i) {
		x = pts[i].x - px;
		y = pts[i].y - py;
		dist = sqrt(x * x + y * y);

		if (dist < GLOB_PICK_AREA)
			return pts + i;
	}

	return NULL;
}

void phys_pick(float px, float py)
{
	concave_t	*curr = all;
	point_t		*ret;

	if (!curr)
		return;

	do {
		do {
			if (curr->flags & CONCAVE_CONST)
				break;

			ret = phys_dist_point(curr, px, py);

			if (ret) {
				glob_pick_point = ret;
				glob_pick = 1;
			}
		}
		while (0);

		curr = curr->next;
	}
	while (curr != all);
}

void phys_pick_move(float px, float py)
{
	glob_pick_dst.x = px;
	glob_pick_dst.y = py;
}

void phys_pick_detach()
{
	glob_pick = 0;
}

